/*
 * Copyright (c) 2000, 2002, 2003 Greg Haerr <greg@censoft.com>
 * Portions Copyright (c) 2002 by Koninklijke Philips Electronics N.V.
 *
 * Freetype TrueType routines originally contributed by Martin Jolicoeur
 * Heavily modified by Greg Haerr
 *
 * Enhanced by Michael Bhme in 2015 to support glyph caching, faster drawing routines
 */
/*#define NDEBUG*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
//#include <dirent.h>
#include <freetype1/freetype.h>
#include <freetype1/extend/ftxkern.h>
#include <freetype1/ftnameid.h>
#include <freetype1/extend/ftxcmap.h>
#include <freetype1/extend/ftxwidth.h>
#include "device.h"
#include "devfont.h"

#if TT_FREETYPE_MAJOR != 1 || TT_FREETYPE_MINOR < 3
#error "You must link with freetype lib version 1.3.x +, and not freetype 2."
#endif

// MBC: Workaround for half and hair space, which are missing in some font files
#define FONT_HALF_SPACE_WORKAROUND		1
// MBC: use fixed witdth for hair space
#define FONT_HAIR_SPACE_WIDTH_OVERRIDE	1

#define UC16_SPACE_CODE				0x0020
//#define UC16_HALF_SPACE_CODE		???
#define UC16_THIN_SPACE_CODE		0x2009
#define UC16_HAIR_SPACE_CODE		0x200A

// width of half and hair space in relation to a normal space char
#define HALF_SPACE_RATIO			2
#define THIN_SPACE_RATIO			3

// fixed multiple of 64 for hair space width
#define HAIR_SPACE_WIDTH			1


// MBC: Cache Extension
#define FONT_CACHE_ENABLE		1
// 256 fixed entries for ASCII characters
#define FONT_CACHE_FIXED_SIZE	256
// additional entries allocated to non-ASCII characters
#define FONT_CACHE_LUT_SIZE		256

//#define FONT_CACHE_BITMAP_SIZE	(1 * 1024 * 1024)	not working

typedef struct MWFREETYPEFONTCACHE
{
	unsigned short	character;	/* != 0 if glyph handle is valid */
	TT_UShort		curchar;	/* char_map lookup result of character */
	TT_Glyph		glyph;		/* glyph handle */
	void			*raster;	/* glyph raster bitmap */
	MWPIXELVAL		*bitmap;	/* bitmap buffer */
	MWPIXELVAL		foreground;	/* last used fg_color */
	MWPIXELVAL		background;	/* last used bg_color */
	TT_F26Dot6		xmin, ymin, xmax, ymax;	/* bounding box */
	TT_Pos			advance;				/* metrics */
	short			cols;		/* glyph raster cols */
	short			width;		/* glyph width */
	short			height;		/* glyph height */
	char			valid;		/* entry is valid */

} MWFREETYPEFONTCACHE;

typedef struct MWFREETYPEFONT {
	PMWFONTPROCS	fontprocs;	/* common hdr*/
	MWCOORD		fontsize;
	int		fontrotation;
	int		fontattr;

	TT_Face		face;		/* freetype stuff*/
	TT_Instance	instance;
	TT_CharMap	char_map;
	TT_Kerning	directory;
	TT_Matrix	matrix;
	TT_Glyph	glyph;
	MWBOOL		can_kern;
	int			last_glyph_code;
	int			last_pen_pos;

#if FONT_CACHE_ENABLE
#warning "FONT_CACHE_ENABLE"

	MWFREETYPEFONTCACHE		*cache;
#ifdef FONT_CACHE_BITMAP_SIZE
	MWPIXELVAL				*bitmaps;
	MWPIXELVAL				*bitmapsCur;
	MWPIXELVAL				*bitmapsEnd;
#endif
#endif

} MWFREETYPEFONT;

int  freetype_init(PSD psd);
PMWFREETYPEFONT freetype_createfont(const char *name, MWCOORD height, int attr);

static MWBOOL freetype_getfontinfo(PMWFONT pfont, PMWFONTINFO pfontinfo);
static void freetype_gettextsize(PMWFONT pfont, const void *text, int cc,
		MWTEXTFLAGS flags, MWCOORD *pwidth, MWCOORD *pheight,
		MWCOORD *pbase);
static void freetype_destroyfont(PMWFONT pfont);
static void freetype_drawtext(PMWFONT pfont, PSD psd, MWCOORD x, MWCOORD y,
		const void *text, int cc, MWTEXTFLAGS flags);
static void freetype_setfontsize(PMWFONT pfont, MWCOORD fontsize);
static void freetype_setfontrotation(PMWFONT pfont, int tenthdegrees);

/* handling routines for MWFREETYPEFONT*/
static MWFONTPROCS freetype_procs = {
	MWTF_UC16,			/* routines expect unicode 16*/
	freetype_getfontinfo,
	freetype_gettextsize,
	NULL,				/* gettextbits*/
	freetype_destroyfont,
	freetype_drawtext,
	freetype_setfontsize,
	freetype_setfontrotation,
	NULL,				/* setfontattr*/
	NULL,				/* duplicate not yet implemented */
};

static TT_Engine 	engine;		/* THE ONLY freetype engine */
//static OUTPIXELVAL gray_palette[5];

/* temp extern decls*/
extern MWPIXELVAL gr_foreground;
extern MWPIXELVAL gr_background;
extern MWBOOL gr_usebg;

int
freetype_init(PSD psd)
{
	static int inited = 0;

	if (inited)
		return 1;

	/* Init freetype library */
	if (TT_Init_FreeType (&engine) != TT_Err_Ok) {
		return 0;
	}

	/* Init kerning extension */
	if (TT_Init_Kerning_Extension (engine) != TT_Err_Ok)
		return 0;

	inited = 1;
	return 1;
}

PMWFREETYPEFONT
freetype_createfont(const char *name, MWCOORD height, int attr)
{
	PMWFREETYPEFONT 	pf;
	unsigned short 		i, n;
	unsigned short 		platform, encoding;
	TT_Face_Properties 	properties;
	char *			p;
	char			fontname[128];

	/* check for pathname prefix*/
	if (strchr(name, '/') != NULL)
		strcpy(fontname, name);
	else {
		strcpy(fontname, FREETYPE_FONT_DIR "/");
		strcat(fontname, name);
	}

	/* check for extension*/
	if ((p = strrchr(fontname, '.')) == NULL || strcmp(p, ".ttf") != 0) {
		strcat(fontname, ".ttf");
	}

	/* allocate font structure*/
	pf = (PMWFREETYPEFONT)GdCalloc(sizeof(MWFREETYPEFONT), 1);
	if (!pf)
		return NULL;
	pf->fontprocs = &freetype_procs;

	/* Load face */
	if (TT_Open_Face (engine, fontname, &pf->face) != TT_Err_Ok)
		goto out;

	/* Load first kerning table */
	pf->can_kern = TRUE;
	if (TT_Load_Kerning_Table (pf->face, 0) != TT_Err_Ok)
		pf->can_kern = FALSE;
	else {
		if (TT_Get_Kerning_Directory (pf->face, &pf->directory)
		    != TT_Err_Ok)
			pf->can_kern = FALSE;
		else {
			/* Support only version 0 kerning table ... */
			if ((pf->directory.version != 0) ||
				(pf->directory.nTables <= 0) ||
				(pf->directory.tables->loaded != 1) ||
				(pf->directory.tables->version != 0) ||
				(pf->directory.tables->t.kern0.nPairs <= 0))
					pf->can_kern = FALSE;
		}
	}

	/* get face properties and allocate preload arrays */
	TT_Get_Face_Properties (pf->face, &properties);

#if 0
	/*
	 * Use header information for ascent and descent
	 * to compute scaled ascent/descent for current font height.
	 */
	h = properties.os2->sTypoAscender - properties.os2->sTypoDescender
		+ properties.os2->sTypoLineGap;
	ascent = properties.os2->sTypoAscender
		+ properties.os2->sTypoLineGap/2;
	pf->ascent = (ascent * height + h/2) / h;
	pf->descent = height - pf->ascent;
#endif
	/* Create a glyph container */
	if (TT_New_Glyph (pf->face, &pf->glyph) != TT_Err_Ok)
		goto out;

	/* create instance */
	if (TT_New_Instance (pf->face, &pf->instance) != TT_Err_Ok)
		goto out;

	/* Set the instance resolution */
	if (TT_Set_Instance_Resolutions (pf->instance, 96, 96) != TT_Err_Ok)
		goto out;

	/* Look for a Unicode charmap: Windows flavor of Apple flavor only */
	n = properties.num_CharMaps;

	for (i = 0; i < n; i++) {
		TT_Get_CharMap_ID (pf->face, i, &platform, &encoding);
		if (((platform == TT_PLATFORM_MICROSOFT) &&
			(encoding == TT_MS_ID_UNICODE_CS)) ||
				((platform == TT_PLATFORM_APPLE_UNICODE) &&
			 		(encoding == TT_APPLE_ID_DEFAULT)))
		{
			TT_Get_CharMap (pf->face, i, &pf->char_map);
			i = n + 1;
		}
	}
	if (i == n) {
		DPRINTF("freetype_createfont: no unicode map table\n");
		goto out;
	}

	GdSetFontSize((PMWFONT)pf, height);
	GdSetFontRotation((PMWFONT)pf, 0);
	GdSetFontAttr((PMWFONT)pf, attr, 0);

#if FONT_CACHE_ENABLE
#ifdef FONT_CACHE_LUT_SIZE
	/* Allocate Glyph Cache */
	pf->cache = (MWFREETYPEFONTCACHE*)GdCalloc(FONT_CACHE_FIXED_SIZE + FONT_CACHE_LUT_SIZE, sizeof(MWFREETYPEFONTCACHE));
#else
	/* Allocate Glyph Cache */
	pf->cache = (MWFREETYPEFONTCACHE*)GdCalloc(FONT_CACHE_FIXED_SIZE, sizeof(MWFREETYPEFONTCACHE));
#endif

#ifdef FONT_CACHE_BITMAP_SIZE
	if (pf->cache != NULL) {
		/* Allocate Glyph Bitmap Buffer */
		pf->bitmaps = (MWPIXELVAL*)GdCalloc(FONT_CACHE_BITMAP_SIZE, sizeof(MWPIXELVAL));
		if (pf->bitmaps != NULL) {
			pf->bitmapsCur = pf->bitmaps;
			pf->bitmapsEnd = pf->bitmaps + FONT_CACHE_BITMAP_SIZE;
		}
	}

	if (pf->bitmaps == NULL) {
		GdFree(pf->cache);
		pf->cache = NULL;
	}
#endif
#endif	// FONT_CACHE_ENABLE

	return pf;

out:
	GdFree2(pf, sizeof(MWFREETYPEFONT));
	return NULL;
}

static int
compute_kernval(PMWFREETYPEFONT pf, TT_UShort current_glyph_code)
{
	int 		i = 0;
	int 		kernval;
	int 		nPairs = pf->directory.tables->t.kern0.nPairs;
	TT_Kern_0_Pair *pair = pf->directory.tables->t.kern0.pairs;

	if (pf->last_glyph_code > -1) {
		TT_UShort last_glyph_code = pf->last_glyph_code;

		while ((pair->left != last_glyph_code)
			&& (pair->right != current_glyph_code))
		{
			pair++;
			i++;
			if (i == nPairs)
				break;
		}

		if (i == nPairs)
			kernval = 0;
		else
			/* We round the value (hence the +32) */
			kernval = (pair->value + 32) & -64;
	} else
		kernval = 0;

	return kernval;
}

#if FONT_HALF_SPACE_WORKAROUND
#warning "FONT_HALF_SPACE_WORKAROUND"

#if FONT_HAIR_SPACE_WIDTH_OVERRIDE
#warning "FONT_HAIR_SPACE_WIDTH_OVERRIDE"
#endif

static int
half_space_check(PMWFREETYPEFONT pf, TT_UShort glyph_index, TT_UShort *curchar)
{
#ifdef UC16_HALF_SPACE_CODE
	if (glyph_index == UC16_HALF_SPACE_CODE)
	{
		*curchar = TT_Char_Index (pf->char_map, UC16_SPACE_CODE);
		// handled
		return 1;
	}
	else
#endif
#ifdef UC16_THIN_SPACE_CODE
	if (glyph_index == UC16_THIN_SPACE_CODE)
	{
		*curchar = TT_Char_Index (pf->char_map, UC16_SPACE_CODE);
		// handled
		return 1;
	}
	else
#endif
#ifdef UC16_HAIR_SPACE_CODE
	if (glyph_index == UC16_HAIR_SPACE_CODE)
	{
		*curchar = TT_Char_Index (pf->char_map, UC16_SPACE_CODE);
		// handled
		return 1;
	}
	else
#endif
	{
		// not handled
		return 0;
	}
}

int half_space_metric(TT_UShort glyph_index, TT_Glyph_Metrics *metrics)
{
#ifdef UC16_HALF_SPACE_CODE
	if (glyph_index == UC16_HALF_SPACE_CODE)
	{
		metrics->advance /= HALF_SPACE_RATIO;
		// minimum width: 64 dots per pixel
		if (metrics->advance < 64)
			metrics->advance = 64;
		// handled
		return 1;
	}
	else
#endif
#ifdef UC16_THIN_SPACE_CODE
	if (glyph_index == UC16_THIN_SPACE_CODE)
	{
		metrics->advance /= THIN_SPACE_RATIO;
		// minimum width: 64 dots per pixel
		if (metrics->advance < 64)
			metrics->advance = 64;
		// handled
		return 1;
	}
	else
#endif
#ifdef UC16_HAIR_SPACE_CODE
	if (glyph_index == UC16_HAIR_SPACE_CODE)
	{
		// minimum width: 64 dots per pixel
		if (metrics->advance > 64 * HAIR_SPACE_WIDTH)
			metrics->advance = 64 * HAIR_SPACE_WIDTH;
		// handled
		return 1;
	}
	else
#endif
	{
		// not handled
		return 0;
	}
}

#endif	// FONT_HALF_SPACE_WORKAROUND


static TT_UShort
Get_Glyph_Width(PMWFREETYPEFONT pf, TT_UShort glyph_index)
{
	TT_UShort curchar;
	TT_Glyph_Metrics metrics;

	/* returns 0 if not found! */
	curchar = TT_Char_Index (pf->char_map,glyph_index);

	if (TT_Load_Glyph ( pf->instance, pf->glyph,
		curchar,
		TTLOAD_SCALE_GLYPH|TTLOAD_HINT_GLYPH) != TT_Err_Ok)
	{
		/* Try to load default glyph: index 0 */
		if (TT_Load_Glyph ( pf->instance, pf->glyph, 0,
			TTLOAD_SCALE_GLYPH|TTLOAD_HINT_GLYPH) != TT_Err_Ok)
			return 0;
	}

	TT_Get_Glyph_Metrics (pf->glyph, &metrics);
	return((metrics.advance & 0xFFFFFFC0) >> 6);
}

#if FONT_CACHE_ENABLE

static void
freetype_clearcache (PMWFREETYPEFONT pf, int freeGlyphs)
{
	int	i, imax;

	if (pf->cache != NULL) {
		imax = FONT_CACHE_FIXED_SIZE;
#ifdef FONT_CACHE_LUT_SIZE
		imax += FONT_CACHE_LUT_SIZE;
#endif
		/* free allocated buffers */
		for (i = 0; i < imax; i++) {
			if (freeGlyphs && pf->cache[i].character != 0) {
				TT_Done_Glyph(pf->cache[i].glyph);
				pf->cache[i].character = 0;
			}
			if (pf->cache[i].raster != NULL) {
				GdFree2(pf->cache[i].raster, pf->cache[i].height * pf->cache[i].cols * 1);
				pf->cache[i].raster = NULL;
			}
			if (pf->cache[i].bitmap != NULL) {
				GdFree2(pf->cache[i].bitmap, pf->cache[i].height * pf->cache[i].width * sizeof(MWPIXELVAL));
				pf->cache[i].bitmap = NULL;
			}
			pf->cache[i].valid = 0;
		}
	}
#ifdef FONT_CACHE_BITMAP_SIZE
	if (pf->bitmaps != NULL)
		pf->bitmapsCur = pf->bitmaps;
#endif
}

void drawchar_points(PSD psd, MWCOORD x, MWCOORD y, MWCOORD width, MWCOORD height, MWPIXELVAL *pixels, MWPIXELVAL bg_color)
{
	MWCOORD count; /* remaining number of cells */
	MWCOORD minx; /* minimum x value */
	MWCOORD maxx; /* maximum x value */
	MWPIXELVAL fg_color; /* pixel foreground color */

	minx = x;
	maxx = x + width - 1;

	/* Set up area clipping, and just return if nothing is visible */
	switch (GdClipArea(psd, minx, y, maxx, y + height - 1))
	{
	case CLIP_VISIBLE:
		/* completely visible */
		count = width * height;
		if (gr_usebg) {
			/* draw with background */
			while (count > 0)
			{
				/* read the pixel value from the pixtype*/
				fg_color = *pixels;
				pixels++;

				psd->DrawPixel(psd, x, y, fg_color);	//drawpoint(psd, x, y);
				if (++x > maxx)
				{
					x = minx;
					y++;
				}
				--count;
			}
		} else {
			/* transparent texts */
			while (count > 0)
			{
				/* read the pixel value from the pixtype*/
				fg_color = *pixels;
				pixels++;

				if (fg_color != bg_color)
					psd->DrawPixel(psd, x, y, fg_color);	//drawpoint(psd, x, y);
				if (++x > maxx)
				{
					x = minx;
					y++;
				}
				--count;
			}
		}
		break;

	case CLIP_INVISIBLE:
		/* invisible */
		return;

	default:
		/* partially visible */
		count = width * height;
		while (count > 0)
		{
			/* read the pixel value from the pixtype*/
			fg_color = *pixels;
			pixels++;

			if (gr_usebg || (fg_color != bg_color))
			{
				if (GdClipPoint(psd, x, y))		// drawpoint(psd, x, y);
					psd->DrawPixel(psd, x, y, fg_color);
			}
			if (++x > maxx)
			{
				x = minx;
				y++;
			}
			--count;
		}
		break;
	}
	GdFixCursor(psd);
}

static void
drawchar_chachebitmap (PMWFREETYPEFONT pf, MWFREETYPEFONTCACHE *entry, OUTPIXELVAL gray_palette[5])
{
	TT_F26Dot6 	xmin, ymin, xmax, ymax, x, y, z, zmax;
	int 		size, width, height;
	MWPIXELVAL	fg_color = entry->foreground;
	MWPIXELVAL	bg_color = entry->background;
	unsigned char 	*src, *srcptr;
	MWPIXELVAL 	*dst, *dstptr;

	/* restore parameters */
	xmin = entry->xmin;
	xmax = entry->xmax;
	ymin = entry->ymin;
	ymax = entry->ymax;
	height = entry->height;
	width  = entry->width;
	size = width * height;

	src = (unsigned char *) entry->raster;
	dst = entry->bitmap + (size - width);

	if (pf->fontattr&MWTF_ANTIALIAS) {
		for (y = ymin; y < ymax; y++) {
			srcptr = src;
			dstptr = dst;

			for (x = xmin; x < xmax; x++) {
				/* fix for transparent texts: unused pixels are drawn in background color */
				if ((int) *srcptr != 0)
					*dstptr++ = gray_palette[(int) *srcptr];
				else
					*dstptr++ = bg_color;
				srcptr++;
			}
			src += entry->cols;
			dst -= width;
		}
	} else {
		for (y = ymin; y < ymax; y++) {
			srcptr = src;
			dstptr = dst;

			for (x = xmin; x < xmax; x++) {
				unsigned int tmp = *srcptr;

				zmax = ((xmax - x - 1) < 7 ? (xmax - x - 1) : 7);
				for (z = 0; z <= zmax; z++) {
					*dstptr++ = ((tmp << z) & 0x80) ? fg_color : bg_color;
				}
				x += 7;
				srcptr++;
			}
			src += entry->cols;
			dst -= width;
		}
	}
}

/* Render a single glyph and add data to cache */
static void
drawchar_addcache(PMWFREETYPEFONT pf, PSD psd, MWFREETYPEFONTCACHE *entry, int x_offset, int y_offset,
	MWPIXELVAL fg_color, MWPIXELVAL bg_color, OUTPIXELVAL gray_palette[5])
{
	TT_F26Dot6	xmin, ymin, xmax, ymax;
	int			size, width, height;
	TT_Outline	outline;
	TT_BBox 	bbox;
	TT_Raster_Map	raster;
	TT_Error	error;
	/*MWPIXELVAL 	save;*/

	/* we begin by grid-fitting the bounding box */
	TT_Get_Glyph_Outline (entry->glyph, &outline);
	TT_Get_Outline_BBox (&outline, &bbox);

	xmin = (bbox.xMin & -64) >> 6;
	ymin = (bbox.yMin & -64) >> 6;
	xmax = ((bbox.xMax + 63) & -64) >> 6;
	ymax = ((bbox.yMax + 63) & -64) >> 6;
	width = xmax - xmin;
	height = ymax - ymin;
	size = width * height;

	/* store parameters */
	entry->xmin = xmin;
	entry->xmax = xmax;
	entry->ymin = ymin;
	entry->ymax = ymax;
	entry->height = height;
	entry->width  = width;

	/* now re-allocate the raster bitmap */
	raster.rows = height;
	raster.width = width;

	if (pf->fontattr&MWTF_ANTIALIAS)
		raster.cols = (raster.width + 3) & -4;	/* pad to 32-bits */
	else
		raster.cols = (raster.width + 7) & -8;	/* pad to 64-bits ??? */
	entry->cols = raster.cols;

	raster.flow = TT_Flow_Up;
	raster.size = raster.rows * raster.cols;

	/* glyph has no bitmap */
	if (raster.size == 0) {
		/* entry is valid */
		entry->valid = 1;
		return;
	}

#ifdef FONT_CACHE_BITMAP_SIZE
//	entry->raster = pf->bitmapsCur;
//	pf->bitmapsCur += raster.size;
#endif

	entry->raster = GdCalloc (raster.size, 1);
	if (!entry->raster)
		return;

	/* now render the glyph in the small pixmap */
	raster.bitmap = entry->raster;

	/* IMPORTANT NOTE: the offset parameters passed to the function     */
	/* TT_Get_Glyph_Bitmap() must be integer pixel values, i.e.,        */
	/* multiples of 64.  HINTING WILL BE RUINED IF THIS ISN'T THE CASE! */
	/* This is why we _did_ grid-fit the bounding box, especially xmin  */
	/* and ymin.                                                        */

	if (!(pf->fontattr&MWTF_ANTIALIAS))
		error = TT_Get_Glyph_Bitmap (entry->glyph, &raster,
				-xmin * 64, -ymin * 64);
	else
		error = TT_Get_Glyph_Pixmap (entry->glyph, &raster,
				-xmin * 64, -ymin * 64);

	if (error) {
		GdFree2 (entry->raster, raster.size);
		entry->raster = NULL;
		return;
	}

	entry->cols = raster.cols;

	entry->bitmap = GdCalloc (size * sizeof (MWPIXELVAL), 1);
	if (!entry->bitmap) {
		GdFree2 (entry->raster, raster.size);
		entry->raster = NULL;
		return;
	}

	entry->foreground = fg_color;
	entry->background = bg_color;

	/* draw the bitmap buffer */
	drawchar_chachebitmap (pf, entry, gray_palette);

	/* FIXME - must clear background upstairs if not gr_usebg*/
	/* FIXME: GdArea problem if fg == bg*/
	/*save = gr_background;*/
	/*gr_background = gr_foreground + 1;*/

	/* Now draw the bitmap ... */
//	GdArea(psd, x_offset + xmin, y_offset - (ymin + height), width, height, entry->bitmap, MWPF_PIXELVAL);
//	GdAreaMWPIXELVAL(psd, x_offset + xmin, y_offset - (ymin + height), width, height, entry->bitmap);
	drawchar_points(psd, x_offset + xmin, y_offset - (ymin + height), width, height, entry->bitmap, entry->background);

	/*gr_background = save;*/

	/* entry is valid */
	entry->valid = 1;
}


/* Render a single glyph directly from cache */
static void
drawchar_cache(PMWFREETYPEFONT pf, PSD psd, MWFREETYPEFONTCACHE *entry, int x_offset, int y_offset,
	MWPIXELVAL fg_color, MWPIXELVAL bg_color, OUTPIXELVAL gray_palette[5])
{
	TT_F26Dot6		xmin, ymin;
	int				size, width, height;
	/*MWPIXELVAL 	save;*/

	/* restore parameters */
	xmin = entry->xmin;
	ymin = entry->ymin;
	height = entry->height;
	width  = entry->width;
	size = width * height;

	if (size > 0) {
		/* redraw if colors changed */
		if (   entry->foreground != fg_color
			|| entry->background != bg_color)
		{
			entry->foreground = fg_color;
			entry->background = bg_color;

			/* draw the bitmap */
			drawchar_chachebitmap (pf, entry, gray_palette);
		}

		/* FIXME - must clear background upstairs if not gr_usebg*/
		/* FIXME: GdArea problem if fg == bg*/
		/*save = gr_background;*/
		/*gr_background = gr_foreground + 1;*/

		/* Now draw the bitmap ... */
	//	GdArea(psd, x_offset + xmin, y_offset - (ymin + height), width, height, entry->bitmap, MWPF_PIXELVAL);
	//	GdAreaMWPIXELVAL(psd, x_offset + xmin, y_offset - (ymin + height), width, height, entry->bitmap);
		drawchar_points(psd, x_offset + xmin, y_offset - (ymin + height), width, height, entry->bitmap, entry->background);

		// aus corefont_drawtext:
	//	if (clip == CLIP_VISIBLE)
	//		drawbitmap(psd, x, y, width, height, bitmap);
	//	else
	//		GdBitmap(psd, x, y, width, height, bitmap);

		/*gr_background = save;*/
	} else {
		/* glyph has no bitmap (space) */
	}
}

/*
 * Draw unicode 16 text string using FREETYPE type font
 */
static void
freetype_drawcache (PMWFREETYPEFONT pf, PSD psd, MWFREETYPEFONTCACHE *entry, TT_F26Dot6 *px, TT_F26Dot6 *py,
	MWPIXELVAL fg_color, MWPIXELVAL bg_color, OUTPIXELVAL gray_palette[5])
{
	TT_F26Dot6 	x = *px, y = *py;
	TT_Pos 		vec_x, vec_y;
	TT_Outline 	outline;
	TT_UShort 	curchar;

	curchar = entry->curchar;

	if (pf->fontrotation) {
		TT_Get_Glyph_Outline (entry->glyph, &outline);
		TT_Transform_Outline (&outline, &pf->matrix);
	}

	if ((pf->fontattr&MWTF_KERNING) && pf->can_kern) {
		vec_x = compute_kernval(pf, curchar);
		if (pf->fontrotation) {
			vec_y = 0;
			TT_Transform_Vector(&vec_x, &vec_y, &pf->matrix);

			x += vec_x / 64;
			y -= vec_y / 64;
		} else
			x += vec_x / 64;
	}

	drawchar_cache(pf, psd, entry, x, y, fg_color, bg_color, gray_palette);

	if (pf->fontrotation) {
		vec_x = entry->advance;
		vec_y = 0;
		TT_Transform_Vector (&vec_x, &vec_y, &pf->matrix);

		x += vec_x / 64;
		y -= vec_y / 64;
	} else {
		x += entry->advance / 64;

		/* Kerning point syndrome avoidance */
		if (pf->last_pen_pos > x)
			x = pf->last_pen_pos;
		pf->last_pen_pos = x;
	}

	pf->last_glyph_code = curchar;

	*px = x;
	*py = y;
}
#endif	// FONT_CACHE_ENABLE

/* Render a single glyph*/
static void
drawchar(PMWFREETYPEFONT pf, PSD psd, TT_Glyph glyph, int x_offset,
	int y_offset, MWPIXELVAL fg_color, MWPIXELVAL bg_color, OUTPIXELVAL gray_palette[5])
{
	TT_F26Dot6 	xmin, ymin, xmax, ymax, x, y, z;
	unsigned char 	*src, *srcptr;
	MWPIXELVAL 	*dst, *dstptr;
	MWPIXELVAL 	*bitmap;
	int 		size, width, height;
	TT_Outline 	outline;
	TT_BBox 	bbox;
	TT_Raster_Map 	Raster;
	TT_Error 	error;
	/*MWPIXELVAL 	save;*/

	/* we begin by grid-fitting the bounding box */
	TT_Get_Glyph_Outline (pf->glyph, &outline);
	TT_Get_Outline_BBox (&outline, &bbox);

	xmin = (bbox.xMin & -64) >> 6;
	ymin = (bbox.yMin & -64) >> 6;
	xmax = ((bbox.xMax + 63) & -64) >> 6;
	ymax = ((bbox.yMax + 63) & -64) >> 6;
	width = xmax - xmin;
	height = ymax - ymin;
	size = width * height;

	/* now re-allocate the raster bitmap */
	Raster.rows = height;
	Raster.width = width;

	if (pf->fontattr&MWTF_ANTIALIAS)
		Raster.cols = (Raster.width + 3) & -4;	/* pad to 32-bits */
	else
		Raster.cols = (Raster.width + 7) & -8;	/* pad to 64-bits ??? */

	Raster.flow = TT_Flow_Up;
	Raster.size = Raster.rows * Raster.cols;
	Raster.bitmap = GdMalloc (Raster.size);
	if (!Raster.bitmap)
		return;

	memset (Raster.bitmap, 0, Raster.size);

	/* now render the glyph in the small pixmap */

	/* IMPORTANT NOTE: the offset parameters passed to the function     */
	/* TT_Get_Glyph_Bitmap() must be integer pixel values, i.e.,        */
	/* multiples of 64.  HINTING WILL BE RUINED IF THIS ISN'T THE CASE! */
	/* This is why we _did_ grid-fit the bounding box, especially xmin  */
	/* and ymin.                                                        */

	if (!(pf->fontattr&MWTF_ANTIALIAS))
		error = TT_Get_Glyph_Bitmap (pf->glyph, &Raster,
				-xmin * 64, -ymin * 64);
	else
		error = TT_Get_Glyph_Pixmap (pf->glyph, &Raster,
				-xmin * 64, -ymin * 64);

	if (error) {
		GdFree2 (Raster.bitmap, Raster.size);
		return;
	}

	bitmap = GdMalloc (size * sizeof (MWPIXELVAL));
	if (!bitmap) {
		GdFree2(Raster.bitmap, Raster.size);
		return;
	}
	memset (bitmap, 0, size * sizeof (MWPIXELVAL));

	src = (unsigned char *) Raster.bitmap;
	dst = bitmap + (size - width);

	for (y = ymin; y < ymax; y++) {
		srcptr = src;
		dstptr = dst;

		for (x = xmin; x < xmax; x++) {
			if (pf->fontattr&MWTF_ANTIALIAS)
				/* fix for transparent texts: unused pixels are drawn in background color */
				if ((int) *srcptr != 0)
					*dstptr++ = gray_palette[(int) *srcptr];
				else
					*dstptr++ = bg_color;
			else {
				for(z=0;
				    z <= ((xmax-x-1) < 7 ? (xmax-x-1) : 7);
				    z++) {
					*dstptr++ = ((*srcptr << z) & 0x80)?
						fg_color: bg_color;
				}
				x += 7;
			}

			srcptr++;
		}

		src += Raster.cols;
		dst -= width;
	}

	/* FIXME - must clear background upstairs if not gr_usebg*/
	/* FIXME: GdArea problem if fg == bg*/
	/*save = gr_background;*/
	/*gr_background = gr_foreground + 1;*/

	/* Now draw the bitmap ... */
	GdArea(psd, x_offset + xmin, y_offset - (ymin + height), width, height,
		bitmap, MWPF_PIXELVAL);

	/*gr_background = save;*/

	GdFree2 (bitmap, size * sizeof (MWPIXELVAL));
	GdFree2 (Raster.bitmap, Raster.size);
}


#if FONT_CACHE_ENABLE
int freetype_drawtext_cache (PMWFREETYPEFONT pf, PSD psd, const unsigned short character, TT_F26Dot6 *px, TT_F26Dot6 *py,
	MWPIXELVAL fg_color, MWPIXELVAL bg_color, OUTPIXELVAL gray_palette[5])
{
	TT_Pos		vec_x, vec_y;
	TT_Outline	outline;
	TT_Glyph_Metrics metrics;

	MWFREETYPEFONTCACHE	*cache = pf->cache;
	MWFREETYPEFONTCACHE	*entry;
	int			handled;
#if FONT_HALF_SPACE_WORKAROUND
	int			space_handled;
#endif
	TT_F26Dot6	x = *px;
	TT_F26Dot6	y = *py;

	entry = NULL;
	handled = 0;

	/* check cache */
	if (cache != NULL && character != 0) {
		if (character < FONT_CACHE_FIXED_SIZE) {
			if (cache[character].valid != 0) {
				/* Bitmap found in cache */
				entry = &cache[character];
				handled = 1;
			} else {
				/* add to cache */
				entry = &cache[character];
			}
		} else {
#ifdef FONT_CACHE_LUT_SIZE
			int	i;

			for (i = FONT_CACHE_FIXED_SIZE; i < (FONT_CACHE_FIXED_SIZE + FONT_CACHE_LUT_SIZE); i++) {
				if (cache[i].character == character && cache[i].valid != 0) {
					/* Bitmap found in cache */
					entry = &cache[i];
					handled = 1;
					break;

				} else if (cache[i].valid == 0) {
					/* cache is filled from the beginning, and no entry is freed alone,
					 * no need to look past the first invalid entry
					 */
					/* allocate this entry to the new character */
					entry = &cache[i];
					break;
				}
			}
#else
			/* not in cache */
#endif	// FONT_CACHE_LUT_SIZE
		}

		if (entry != NULL) {
			if (handled != 0) {
				/* copy bitmap from cache */
				freetype_drawcache (pf, psd, entry, &x, &y,
					fg_color, bg_color, gray_palette);

			} else {
				/* allocate a new glyph handle if neccessary */
				if (entry->character != 0 || TT_New_Glyph (pf->face, &entry->glyph) == TT_Err_Ok) {

					/* add character to cache */
					entry->character = character;

					/* TT_Char_Index returns 0 if not found! */
					entry->curchar = TT_Char_Index (pf->char_map, character);

#if FONT_HALF_SPACE_WORKAROUND
					if (entry->curchar == 0)
					{
						space_handled = half_space_check (pf, character, &entry->curchar);
					}
#if FONT_HAIR_SPACE_WIDTH_OVERRIDE
					else if (character == UC16_HAIR_SPACE_CODE)
					{
						/* force width of hair space */
						space_handled = 1;
					}
#endif
					else
						space_handled = 0;
#endif

					if (TT_Load_Glyph (pf->instance, entry->glyph, entry->curchar,
						TTLOAD_DEFAULT) != TT_Err_Ok)
					{
						/* failed to load glyph */
					}
					else
					{
						if (pf->fontrotation) {
							TT_Get_Glyph_Outline (entry->glyph, &outline);
							TT_Transform_Outline (&outline, &pf->matrix);
						}

						TT_Get_Glyph_Metrics (entry->glyph, &metrics);

#if FONT_HALF_SPACE_WORKAROUND
						if (space_handled != 0)
							half_space_metric (character, &metrics);
#endif

						entry->advance = metrics.advance;

						if ((pf->fontattr&MWTF_KERNING) && pf->can_kern) {
							vec_x = compute_kernval(pf, entry->curchar);
							if (pf->fontrotation) {
								vec_y = 0;
								TT_Transform_Vector(&vec_x, &vec_y, &pf->matrix);

								x += vec_x / 64;
								y -= vec_y / 64;
							} else
								x += vec_x / 64;
						}

						/* draw and add bitmap to cache */
						drawchar_addcache(pf, psd, entry, x, y,
							fg_color, bg_color, gray_palette);

						if (pf->fontrotation) {
							vec_x = metrics.advance;
							vec_y = 0;
							TT_Transform_Vector (&vec_x, &vec_y, &pf->matrix);

							x += vec_x / 64;
							y -= vec_y / 64;
						} else {
							x += metrics.advance / 64;

							/* Kerning point syndrome avoidance */
							if (pf->last_pen_pos > x)
								x = pf->last_pen_pos;
							pf->last_pen_pos = x;
						}

						pf->last_glyph_code = entry->curchar;

						/* next character */
						handled = 1;
					}
				}
			}
			*px = x;
			*py = y;
		} else {
			/* character not found and no free entry available */
		}
	}
	return handled;
}
#endif	// FONT_CACHE_ENABLE

/*
 * Draw unicode 16 text string using FREETYPE type font
 */
static void
freetype_drawtext(PMWFONT pfont, PSD psd, MWCOORD ax, MWCOORD ay,
	const void *text, int cc, MWTEXTFLAGS flags)
{
	PMWFREETYPEFONT	pf = (PMWFREETYPEFONT)pfont;
	const unsigned short *	str = text;
	TT_F26Dot6 	x = ax, y = ay;
	TT_Pos 		vec_x, vec_y;
	int 		i;
	TT_F26Dot6	startx, starty;
	TT_Outline 	outline;
	TT_UShort 	curchar;
	TT_Glyph_Metrics metrics;
	static unsigned char blend[5] = { 0x00, 0x44, 0x88, 0xcc, 0xff };
	static unsigned char virtual_palette[5] = { 0, 1, 2, 3, 4 };
	OUTPIXELVAL gray_palette[5] = { 0x00, 0x00, 0x00, 0x00, 0x00 };
	MWPIXELVAL fg_color = gr_foreground;
	MWPIXELVAL bg_color = gr_background;

	pf->last_glyph_code = -1;		/* reset kerning*/
	pf->last_pen_pos = -32767;

	if (flags&(MWTF_BOTTOM | MWTF_TOP))
	{
		TT_Face_Properties properties;
		TT_Instance_Metrics imetrics;
		TT_F26Dot6 ascent, descent;

		/*
		 * Compute instance ascent & descent values
		 * in fractional units (1/64th pixel)
		 */
		TT_Get_Face_Properties (pf->face, &properties);
		TT_Get_Instance_Metrics(pf->instance, &imetrics);

		/*
		 * Offset the starting point if necessary,
		 * FreeType always aligns at baseline
		 */
		if (flags&MWTF_BOTTOM) {
			descent = ((properties.horizontal->Descender * imetrics.y_scale)/0x10000);
			vec_x = 0;
			vec_y = descent;
			TT_Transform_Vector(&vec_x, &vec_y, &pf->matrix);
			x -= vec_x / 64;
			y += vec_y / 64;
		} else if (flags&MWTF_TOP) {
			ascent = ((properties.horizontal->Ascender * imetrics.y_scale)/0x10000);
			vec_x = 0;
			vec_y = ascent;
			TT_Transform_Vector(&vec_x, &vec_y, &pf->matrix);
			x -= vec_x / 64;
			y += vec_y / 64;
		}
	}

	/* Set the "graylevels" */
	if (pf->fontattr&MWTF_ANTIALIAS) {
		TT_Set_Raster_Gray_Palette (engine, virtual_palette);

		if (!gr_usebg) {
			/* transparent texts fade to black,
			 * because the actual background color may be different for every pixel...
			 */
			alphablend(psd, gray_palette, fg_color, 0,
				blend, 5);
			/* use max index as transparent color */
			bg_color = ~0;
		} else {
			alphablend(psd, gray_palette, fg_color, bg_color,
				blend, 5);
		}
	} else {
	}

	startx = x;
	starty = y;
	for (i = 0; i < cc; i++) {
#if FONT_HALF_SPACE_WORKAROUND
		int space_handled;
#endif

#if FONT_CACHE_ENABLE
		/* check cache */
		if (freetype_drawtext_cache (pf, psd, str[i], &x, &y, fg_color, bg_color, gray_palette) != 0) {
			/* next character */
			continue;
		}
#endif	// FONT_CACHE_ENABLE

		/* returns 0 if not found! */
		curchar = TT_Char_Index (pf->char_map, str[i]);

#if FONT_HALF_SPACE_WORKAROUND
		if (curchar == 0)
		{
			space_handled = half_space_check (pf, str[i], &curchar);
		}
#if FONT_HAIR_SPACE_WIDTH_OVERRIDE
		else if (str[i] == UC16_HAIR_SPACE_CODE)
		{
			/* force width of hair space */
			space_handled = 1;
		}
#endif
		else
			space_handled = 0;
#endif

		if (TT_Load_Glyph (pf->instance, pf->glyph, curchar,
			TTLOAD_DEFAULT) != TT_Err_Ok)
				continue;

		if (pf->fontrotation) {
			TT_Get_Glyph_Outline (pf->glyph, &outline);
			TT_Transform_Outline (&outline, &pf->matrix);
		}

		TT_Get_Glyph_Metrics (pf->glyph, &metrics);

#if FONT_HALF_SPACE_WORKAROUND
		if (space_handled != 0)
			half_space_metric (str[i], &metrics);
#endif

		if ((pf->fontattr&MWTF_KERNING) && pf->can_kern) {
			if (pf->fontrotation) {
				vec_x = compute_kernval(pf, curchar);
				vec_y = 0;
				TT_Transform_Vector(&vec_x, &vec_y,&pf->matrix);

				x += vec_x / 64;
				y -= vec_y / 64;
			} else
				x += compute_kernval(pf, curchar) / 64;
		}

		drawchar(pf, psd, pf->glyph, x, y, gr_foreground, gr_background, gray_palette);

		if (pf->fontrotation) {
			vec_x = metrics.advance;
			vec_y = 0;
			TT_Transform_Vector (&vec_x, &vec_y, &pf->matrix);

			x += vec_x / 64;
			y -= vec_y / 64;
		} else {
			x += metrics.advance / 64;

			/* Kerning point syndrome avoidance */
			if (pf->last_pen_pos > x)
				x = pf->last_pen_pos;
			pf->last_pen_pos = x;
		}

		pf->last_glyph_code = curchar;
	}

	if (pf->fontattr & MWTF_UNDERLINE)
		GdLine(psd, startx, starty, x, y, FALSE);

	GdFixCursor(psd);
}

/*
 * Return information about a specified font.
 */
static MWBOOL
freetype_getfontinfo(PMWFONT pfont, PMWFONTINFO pfontinfo)
{
	int	i;
	PMWFREETYPEFONT	pf = (PMWFREETYPEFONT)pfont;
	TT_Face_Properties 	properties;
    TT_Instance_Metrics imetrics;
    TT_UShort last_glyph_index;

	TT_Get_Face_Properties (pf->face, &properties);
    TT_Get_Instance_Metrics(pf->instance, &imetrics);

    /* Fill up the fields */
	pfontinfo->height = (((properties.horizontal->Ascender * \
	                        imetrics.y_scale)/ 0x10000) >> 6) -
	                    (((properties.horizontal->Descender * \
	                        imetrics.y_scale)/ 0x10000) >> 6);
	pfontinfo->maxwidth = ((properties.horizontal->xMax_Extent * \
	                        imetrics.x_scale)/ 0x10000) >> 6;
	pfontinfo->baseline = ((properties.horizontal->Ascender * \
	                        imetrics.y_scale)/ 0x10000) >> 6;
	pfontinfo->firstchar = TT_CharMap_First(pf->char_map, NULL);
	pfontinfo->lastchar = TT_CharMap_Last(pf->char_map, NULL);
	pfontinfo->fixed = properties.postscript->isFixedPitch;

	/* FIXME: Wild guesses - should be possible to get these properly
	 * from the library. */
	pfontinfo->linespacing = pfontinfo->height + 2;
	pfontinfo->descent = pfontinfo->height - pfontinfo->baseline;
	pfontinfo->maxascent = pfontinfo->baseline;
	pfontinfo->maxdescent = pfontinfo->descent;

    last_glyph_index = properties.num_Glyphs > 255 ? 255: properties.num_Glyphs-1;

    /* Doesn't work ... don't know why ....*/
#if 0
    if (TT_Get_Face_Widths( pf->face, 0,
                            last_glyph_index, widths, NULL ) != TT_Err_Ok) {
        return TRUE;
    }

    for(i=0; i<=last_glyph_index; i++)
		DPRINTF("widths[%d]: %d\n", i, widths[i]);
#endif

    /* Get glyphs widths */
	for(i=0; i<=last_glyph_index; i++)
		pfontinfo->widths[i] = Get_Glyph_Width(pf, i);

#if 0
    DPRINTF("x_ppem: %d\ny_ppem: %d\nx_scale: %d\ny_scale: %d\n\
    x_resolution: %d\ny_resolution: %d\n",
    imetrics.x_ppem, imetrics.y_ppem, imetrics.x_scale, \
    imetrics.y_scale, imetrics.x_resolution, imetrics.y_resolution);

    DPRINTF("Ascender: %d\nDescender: %d\nxMax_Extent: %d\n\
    Mac Style Say Italic?: %d\nMac Style Say Bold?: %d\n\
    sTypoAscender: %d\nsTypoDescender: %d\nusWinAscent: %d\n\
    usWinDescent: %d\nusFirstCharIndex: %d\nusLastCharIndex: %d\n\
    OS2 Say Italic?: %d\nOS2 Say Bold?: %d\nOS2 Say monospaced?: %d\n\
    Postscript Say monospaced?: %d\n",\
    properties.horizontal->Ascender,
    properties.horizontal->Descender,
    properties.horizontal->xMax_Extent,
    (properties.header->Mac_Style & 0x2)?1:0,
    (properties.header->Mac_Style & 0x1)?1:0,
    properties.os2->sTypoAscender,
    properties.os2->sTypoDescender,
    properties.os2->usWinAscent,
    properties.os2->usWinDescent,
    properties.os2->usFirstCharIndex,
    properties.os2->usLastCharIndex,
    (properties.os2->fsSelection & 0x1)?1:0,
    (properties.os2->fsSelection & 0x20)?1:0,
    properties.postscript->isFixedPitch,
    (properties.os2->panose[3] == 9)?1:0);
#endif

	return TRUE;
}

#if FONT_CACHE_ENABLE
int freetype_gettextsize_cache (PMWFREETYPEFONT pf, const unsigned short character, TT_F26Dot6 *px)
{
	MWFREETYPEFONTCACHE	*cache = pf->cache;
	MWFREETYPEFONTCACHE	*entry = NULL;
	int			handled = 0;
	TT_F26Dot6	x = *px;

	/* check cache */
	if (cache != NULL && character != 0) {
		if (character < FONT_CACHE_FIXED_SIZE) {
			if (cache[character].valid != 0) {
				/* Bitmap found in cache */
				entry = &cache[character];
			}
		} else {
#ifdef FONT_CACHE_LUT_SIZE
			int i;

			for (i = FONT_CACHE_FIXED_SIZE; i < (FONT_CACHE_FIXED_SIZE + FONT_CACHE_LUT_SIZE); i++) {
				if (cache[i].character == character && cache[i].valid != 0) {
					/* Bitmap found in cache */
					entry = &cache[i];
					break;

				} else if (cache[i].valid == 0) {
					/* cache is filled from the beginning, and no entry is freed alone,
					 * no need to look past the first invalid entry
					 */
					break;
				}
			}
#else
			/* not in cache */
#endif	// FONT_CACHE_LUT_SIZE
		}

		if (entry != NULL) {

			if ((pf->fontattr&MWTF_KERNING) && pf->can_kern) {
				x += compute_kernval(pf, entry->curchar) / 64;
			}

			x += entry->advance / 64;

			/* Kerning point syndrome avoidance */
			if (pf->last_pen_pos > x)
				x = pf->last_pen_pos;
			pf->last_pen_pos = x;

			pf->last_glyph_code = entry->curchar;

			/* next character */
			handled = 1;
		}
	}

	*px = x;
	return handled;
}
#endif	// FONT_CACHE_ENABLE

static void
freetype_gettextsize(PMWFONT pfont, const void *text, int cc, MWTEXTFLAGS flags,
	MWCOORD *pwidth, MWCOORD *pheight, MWCOORD *pbase)
{
	PMWFREETYPEFONT	pf = (PMWFREETYPEFONT)pfont;
	const unsigned short *	str = text;
	TT_F26Dot6 	x = 0;
	int 		i;
	TT_UShort 	curchar;
	TT_Glyph_Metrics metrics;
	TT_Face_Properties 	properties;
	TT_Instance_Metrics imetrics;

#if FONT_HALF_SPACE_WORKAROUND
	int space_handled;
#endif

	TT_Get_Face_Properties (pf->face, &properties);
	TT_Get_Instance_Metrics(pf->instance, &imetrics);

	pf->last_glyph_code = -1;		/* reset kerning*/
	pf->last_pen_pos = -32767;

	for (i = 0; i < cc; i++) {

#if FONT_CACHE_ENABLE
		/* check cache */
		if (freetype_gettextsize_cache(pf, str[i], &x) != 0)
		{
			/* next character */
			continue;
		}
#endif	// FONT_CACHE_ENABLE

		/* returns 0 if not found! */
		curchar = TT_Char_Index (pf->char_map, str[i]);

#if FONT_HALF_SPACE_WORKAROUND
		if (curchar == 0)
		{
			space_handled = half_space_check (pf, str[i], &curchar);
		}
#if FONT_HAIR_SPACE_WIDTH_OVERRIDE
		else if (str[i] == UC16_HAIR_SPACE_CODE)
		{
			/* force width of hair space */
			space_handled = 1;
		}
#endif
		else
			space_handled = 0;
#endif

		if (TT_Load_Glyph (pf->instance, pf->glyph, curchar,
			TTLOAD_SCALE_GLYPH|TTLOAD_HINT_GLYPH) != TT_Err_Ok)
				continue;

		TT_Get_Glyph_Metrics (pf->glyph, &metrics);

#if FONT_HALF_SPACE_WORKAROUND
		if (space_handled != 0)
			half_space_metric (str[i], &metrics);
#endif

		if ((pf->fontattr&MWTF_KERNING) && pf->can_kern) {
				x += compute_kernval(pf, curchar) / 64;
		}

		x += metrics.advance / 64;

		/* Kerning point syndrome avoidance */
		if (pf->last_pen_pos > x)
			x = pf->last_pen_pos;
		pf->last_pen_pos = x;

		pf->last_glyph_code = curchar;
	}

	*pwidth = x;
	*pheight = (((properties.horizontal->Ascender * imetrics.y_scale)/ 0x10000) >> 6) - (((properties.horizontal->Descender * imetrics.y_scale)/ 0x10000) >> 6);
	/* FIXME: is it what's required ?? */	// MBC: No, it isn't ;-)
//	*pbase = (((-properties.horizontal->Descender) * imetrics.y_scale)/ 0x10000) >> 6;
	/* cf freetype_getfontinfo */
	*pbase = ((properties.horizontal->Ascender * imetrics.y_scale)/ 0x10000) >> 6;
}

static void
freetype_destroyfont(PMWFONT pfont)
{
	PMWFREETYPEFONT	pf = (PMWFREETYPEFONT)pfont;

	TT_Close_Face(pf->face);

#if FONT_CACHE_ENABLE
	/* clear the cache */
	freetype_clearcache (pf, 1);
	/* free the cache */
	GdFree2(pf->cache, FONT_CACHE_FIXED_SIZE + FONT_CACHE_LUT_SIZE * sizeof(MWFREETYPEFONTCACHE));
#ifdef FONT_CACHE_BITMAP_SIZE
	GdFree(pf->bitmaps);
#endif
#endif
	GdFree2(pf, sizeof(MWFREETYPEFONT));
}

static void
freetype_setfontsize(PMWFONT pfont, MWCOORD fontsize)
{
	PMWFREETYPEFONT	pf = (PMWFREETYPEFONT)pfont;

	pf->fontsize = fontsize;

	/* allow create font with height=0*/
	if (!fontsize)
		return;

	/* We want real pixel sizes ... not points ...*/
	TT_Set_Instance_PixelSizes( pf->instance, pf->fontsize,
                                pf->fontsize, pf->fontsize * 64 );
#if 0
	/* set charsize (convert to points for freetype)*/
	TT_Set_Instance_CharSize (pf->instance,
		((pf->fontsize * 72 + 96/2) / 96) * 64);
#endif

#if FONT_CACHE_ENABLE
	/* clear the cache */
	freetype_clearcache (pf, 0);
#endif	// FONT_CACHE_ENABLE
}

static void
freetype_setfontrotation(PMWFONT pfont, int tenthdegrees)
{
	PMWFREETYPEFONT	pf = (PMWFREETYPEFONT)pfont;
	float 		angle;

	pf->fontrotation = tenthdegrees;

	/* Build the rotation matrix with the given angle */
	TT_Set_Instance_Transform_Flags (pf->instance, TRUE, FALSE);

	angle = pf->fontrotation * M_PI / 1800;
	pf->matrix.yy = (TT_Fixed) (cos (angle) * (1 << 16));
	pf->matrix.yx = (TT_Fixed) (sin (angle) * (1 << 16));
	pf->matrix.xx = pf->matrix.yy;
	pf->matrix.xy = -pf->matrix.yx;

#if FONT_CACHE_ENABLE
	/* clear the cache */
	freetype_clearcache (pf, 0);
#endif	// FONT_CACHE_ENABLE
}

/* FIXME: this routine should work for all font renderers...*/
int
GdGetTextSizeEx(PMWFONT pfont, const void *str, int cc,int nMaxExtent,
	int* lpnFit, int* alpDx,MWCOORD *pwidth,MWCOORD *pheight,
	MWCOORD *pbase, MWTEXTFLAGS flags)
{
	unsigned short	buf[UC16_CONVERT_BUFFER];
	unsigned short* text;
	PMWFREETYPEFONT	pf = (PMWFREETYPEFONT)pfont;
	MWTEXTFLAGS	defencoding = pf->fontprocs->encoding;
	int 		x = 0;
	int 		i;
	TT_UShort 	curchar;
	TT_Glyph_Metrics metrics;
	TT_Face_Properties 	properties;
	TT_Instance_Metrics imetrics;

	if ((cc<0)||(!str))
	{
		*pwidth = *pheight = *pbase = 0;
		return 0;
	}
	/* convert encoding if required*/
	if((flags & (MWTF_PACKMASK|MWTF_DBCSMASK)) != defencoding)
	{
		cc = GdConvertEncoding(str, flags, cc, buf, sizeof(buf), defencoding);
		flags &= ~(MWTF_PACKMASK|MWTF_DBCSMASK);
		flags |= defencoding;
		text=buf;
	} else text =(unsigned short*)str;
	if(cc <= 0)
	{
		*pwidth = *pheight = *pbase = 0;
		return 0;
	}

	TT_Get_Face_Properties (pf->face, &properties);
	TT_Get_Instance_Metrics(pf->instance, &imetrics);

	pf->last_glyph_code = -1;		/* reset kerning*/
	pf->last_pen_pos = -32767;
	if (lpnFit)
		*lpnFit=-1;
	for (i = 0; i < cc; i++)
	{
		/* returns 0 if not found! */
		curchar = TT_Char_Index (pf->char_map,text[i]);

		if (TT_Load_Glyph (pf->instance, pf->glyph, curchar,
			TTLOAD_SCALE_GLYPH|TTLOAD_HINT_GLYPH) != TT_Err_Ok)
		{
//			printf("Unable to load glyph with index=%d\n",curchar);
			return 0;
		}
		TT_Get_Glyph_Metrics (pf->glyph, &metrics);
		if ((pf->fontattr&MWTF_KERNING) && pf->can_kern)
		{
			x += compute_kernval(pf, curchar) / 64;
		}
		x += metrics.advance / 64;
			if((lpnFit)&&(alpDx))
		{
			if (x<=nMaxExtent)
			     alpDx[i]=x;
			else
			     if (*lpnFit==-1)
					(*lpnFit)=i;
		}
		/* Kerning point syndrome avoidance */
		if (pf->last_pen_pos > x)
			x = pf->last_pen_pos;
		pf->last_pen_pos = x;
		pf->last_glyph_code = curchar;
	}
	if ((lpnFit)&&(*lpnFit==-1))
		*lpnFit=cc;
	*pwidth = x;
	*pheight = (((properties.horizontal->Ascender *
	                imetrics.y_scale)/ 0x10000) >> 6) -
			    (((properties.horizontal->Descender *
			    imetrics.y_scale)/ 0x10000) >> 6);
	/* FIXME: is it what's required ??*/
	if (pbase)
		*pbase = (((-properties.horizontal->Descender) *
			    imetrics.y_scale)/ 0x10000) >> 6;
	return 1;
}

#if 0	// MBC: unused
/*
 * This function is taken almost verbatim from ftdump.c from
 * the freetype library (version 1.3.1)
 */
static char *
tt_lookup_name(TT_Face face)
{
	TT_Face_Properties prop;
	unsigned short i, n;
	unsigned short platform, encoding, language, id;
	char *string;
	char *name_buffer;
	unsigned short string_len;
	int j, found;
	int index = 4; /* I dont know why as yet.. */
	int name_len;


	TT_Get_Face_Properties(face, &prop);
	n = prop.num_Names;

	for ( i = 0; i < n; i++ ) {
		TT_Get_Name_ID( face, i, &platform, &encoding, &language, &id );
		TT_Get_Name_String( face, i, &string, &string_len );

		if (id == index ) {
			/* The following code was inspired from Mark Leisher's */
			/* ttf2bdf package                                     */
			found = 0;

			/* Try to find a Microsoft English name */
			if ( platform == 3 )
				for ( j = 1; j >= 0; j-- )
					if ( encoding == j )  /* Microsoft ? */
						if ( (language & 0x3FF) == 0x009 ) {
							/* English language */
							found = 1;
							break;
						}

			if ( !found && platform == 0 && language == 0 )
				found = 1;

			/* Found a Unicode Name. */
			if ( found ) {
				if ( string_len > 512 )
					string_len = 512;

				name_len = 0;
				name_buffer = (char*)GdMalloc((string_len / 2) + 1);

				for ( i = 1; i < string_len; i += 2 )
					name_buffer[name_len++] = string[i];

				name_buffer[name_len] = '\0';

				return name_buffer;
			}
		}
	}

	/* Not found */
	return NULL;
}

static char *
get_tt_name(char *p)
{
	TT_Face face;
	char *ret;

	/*printf("Trying to open: %s!\n",p);*/

	if (TT_Open_Face(engine, p, &face) != TT_Err_Ok) {
//		printf("Error opening font: %s\n", p);
		return NULL;
	}

	ret = tt_lookup_name(face);

	TT_Close_Face(face);

	return ret;
}
#endif

void
GdFreeFontList(MWFONTLIST ***fonts, int n)
{
	int i;
	MWFONTLIST *g, **list = *fonts;

	for (i = 0; i < n; i++) {
		g = list[i];
		if(g) {
			if(g->mwname)
				GdFree(g->mwname);
			if(g->ttname)
				GdFree(g->ttname);
			GdFree(g);
		}
	}
	GdFree(list);
	*fonts = 0;
}

void
GdGetFontList(MWFONTLIST ***fonts, int *numfonts)
{
#if 0	// MBC: not supported
	DIR *dir;
	struct dirent *dent;
	char *p, *ftmp;
	int pl, idx = 0;
	MWFONTLIST **list;


	if (TT_Err_Ok != TT_Init_FreeType(&engine)) {
		printf("Unable to initialize freetype\n");
		*numfonts = -1;
		return ;
	}

	dir = opendir(FREETYPE_FONT_DIR);

	if (dir <= 0) {
		printf("Error opening font directory\n");
		*numfonts = -1;
		return ;
	}

	/* get the number of strings we need to allocate */
	while ((dent = readdir(dir)) != NULL) {
		p = strrchr(dent->d_name, '.');
		if (strcasecmp(p, ".ttf") == 0)
			idx++;
	}

	*numfonts = idx;
	rewinddir(dir);

	/* allocate strings */
	list = (MWFONTLIST**)GdMalloc(idx * sizeof(MWFONTLIST*));
	for (pl = 0; pl < idx; pl++)
		list[pl] = (MWFONTLIST*)GdMalloc(sizeof(MWFONTLIST));

	*fonts = list;

	idx = 0;

	while ((dent = readdir(dir)) != NULL) {
		/* check extension */
		p = strrchr(dent->d_name, '.');

		if (strcasecmp(p, ".ttf") == 0) {

			/* get full path */
			p = 0;
			pl = strlen(FREETYPE_FONT_DIR) + strlen(dent->d_name) *
						sizeof(char) + 2;
			p = (char*)GdMalloc(pl);
			p = (char*)memset(p, '\0', pl);
			p = (char*)strcat(p, FREETYPE_FONT_DIR);
			p = (char*)strcat(p, "/");
			p = (char*)strcat(p, dent->d_name);


			if((ftmp = get_tt_name(p)) != NULL) {
				list[idx]->ttname = ftmp;
				list[idx]->mwname = GdMalloc(strlen(dent->d_name) + 1);
				list[idx]->mwname = strcpy(list[idx]->mwname, dent->d_name);

				idx++;
			}

			GdFree(p);
		}
	}

	closedir(dir);
#else
	printf("Not supported yet: freetype\n");
	*numfonts = -1;
	return;
#endif
}
